#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>

#include <qmsgbox.h>
#include <qstring.h>
#include <kapp.h>
#include "hexbuffer.h"

CHexAction::CHexAction( HexAction action, uint offset )
{
  mAction   = action;
  mOffset   = offset;
  mSize     = 0;
  mData     = 0;
  mDataSize = 0;
  mNext     = 0;
}

CHexAction::~CHexAction( void )
{
  delete [] mData;
}
 
void CHexAction::setData( uint size, char *data, uint dataSize )
{

  if( data != 0 && dataSize > 0 ) 
  {
    mData = new char[ dataSize ];
    if( mData == 0 )
    {
      return;
    }
    memcpy( mData, data, dataSize );
    mDataSize = dataSize;
  } 
  else 
  {
    mDataSize = 0;
    mData = 0;
  }
  mSize = size;

}


CHexActionGroup::CHexActionGroup( uint startOffset, uint startBit )
{
  mStartOffset = startOffset;
  mStartBit    = startBit;
  mHexAction   = 0;
}

CHexActionGroup::~CHexActionGroup( void )
{
  CHexAction *ptr = mHexAction;
  while( ptr != 0 )
  {
    CHexAction *next = ptr->mNext;
    delete ptr;
    ptr = next;
  }
}
 
void CHexActionGroup::insertAction( CHexAction *hexAction )
{
  hexAction->mNext = mHexAction;
  mHexAction = hexAction;
}






CHexBuffer::CHexBuffer( void ) 
  :QByteArray()
{
  mColorIndex = 0;
  mPrintBuf   = 0;
  mActiveEditArea   = edit_primary;
  mDocumentSize     = 0;
  mDocumentModified = false;

  SDisplayState state;
  int errCode = setDisplayState( state );
  if( errCode != 0 )
  {
    return;
  }

  setForegroundColor( black, blue, black );
  setBackgroundColor( white );
  setCursorColor( white, red, black, yellow );
  setSelectionColor( white, darkBlue );
  setSeparatorColor( darkGreen, darkGreen );
  setOffsetColor( red );
  setNonPrintChar( '.' );
  setFont( "courier", 14, QFont::Normal, FALSE );
  setShowCursor( true );
  setDisableCursor( false );
  setEditMode( edit_insert );

  mUndoLimit = 10;
  mUndoIndex = 0;
  mUndoList.setAutoDelete( TRUE );    

  selectionSet( 0, true );
}


CHexBuffer::~CHexBuffer( void ) 
{
  debug("CHexBuffer::~CHexBuffer");
  delete [] mColorIndex;
  delete [] mPrintBuf;
}



int CHexBuffer::setDisplayState( SDisplayState &state )
{
  mDpyState = state;
  mDpyState.verify();

  if( mDpyState.viewMode == SDisplayState::view_ascii )
  {
    mDpyState.textMode = SDisplayState::text_hide;
    mActiveEditArea = edit_primary;
    setEditMode( mEditMode );
  }

  mCursor.setLineSize( mDpyState.lineSize );
  mCursor.addOffset( 0 ); // This will only reset the cell position

  computeLineWidth();
  cursorCompute();

  delete [] mColorIndex; mColorIndex = 0;
  delete [] mPrintBuf; mPrintBuf = 0;

  mColorIndex = new char[ mDpyState.lineSize ];
  if( mColorIndex == 0 )
  {
    return( ENOMEM );
  }
  setForegroundColor( mTextPrimaryFg[0], mTextPrimaryFg[1], mTextSecondaryFg );

  //
  // The 'mPrintBuf' is used to store formatted text. It is used for all 
  // print operations and must have the size of the 'mDpyState.lineSize' which
  // is the number of bytes in one single display line.
  //
  mPrintBuf = new char[ mDpyState.lineSize < 12 ? 12 : mDpyState.lineSize ];
  if( mPrintBuf == 0 )
  {
    delete [] mColorIndex; mColorIndex = 0;
    return( ENOMEM );
  }
  
  return( 0 );
}






bool CHexBuffer::matchWidth( uint width )
{
  if( size() == 0 || (uint)mFixedWidth >= width )
  {
    return( false );
  }

  width -= mFixedWidth;

  uint  g = mDpyState.splitGroup == true ? mDpyState.groupSize : 1;
  uint  n = g * mNumCell;
  uint  u = mUnitWidth;
  uint  s = mDpyState.textMode == SDisplayState::text_hide ? 0 : g;
  uint  o = mDpyState.splitGroup == true ? mSplitWidth : 0;
  float x = (float)(width+o)/(float)(u*(n+s)+o);

  uint lineSize = (uint)x * g;

  if( mDpyState.wrapInGroup == true )
  {
    //
    // Examine if we can add one or more columns from the next group. This
    // will make the rightmost group smaller than the rest but we will 
    // utilize as much of the available space (ie., width) as possible. 
    // (Note that the column itself (which reprsents one byte of filedata) 
    // can not be splitted, eg., in binary mode the column is eight byte 
    // wide and will not be splitted).  
    //
    int w = (int)((float)((int)x)* (float)(u*(n+s)+o) - (float)o);
    if( w > 0 && (uint)w < width )
    {
      width -= w;
      if( width > o )
      {
	x = (float)(width-o) / (float)(u*(mNumCell+1));
	lineSize += (uint)x;
      }
    }
  }
  
  if( lineSize == 0 || lineSize == mDpyState.lineSize )
  {
    //
    // We have to redraw all text if a change occurs so we avoid it if
    // possible.
    //
    return( false );
  }
 
  mDpyState.lineSize  = lineSize;
  setDisplayState( mDpyState );
  return( true );
}







void CHexBuffer::setForegroundColor( const QColor &fg1, const QColor &fg2,
				     const QColor &fg3 )
{
  if( mColorIndex != 0 )
  {
    uint groupSize = mDpyState.groupSize;

    if( groupSize == 0 ) { groupSize = 1; }
    for( uint i = 0, entry = 0;  i < mDpyState.lineSize; i++ )
    {
      if( i > 0 && i % groupSize == 0 ) { entry = entry == 0 ? 1 : 0; }
      mColorIndex[i] = entry;
    }
  }
  mTextPrimaryFg[0] = fg1;
  mTextPrimaryFg[1] = fg2;
  mTextSecondaryFg  = fg3;
}

void CHexBuffer::setSelectionColor( const QColor &fg, const QColor &bg )
{
  mSelectFg = fg;
  mSelectBg = bg;
}

void CHexBuffer::setBackgroundColor( const QColor &color )
{
  mTextBg = color;
}


void CHexBuffer::setCursorColor( const QColor &actFg, const QColor &actBg,
				 const QColor &pasFg, const QColor &pasBg )
{
  mCursorActFg = actFg;
  mCursorActBg = actBg;
  mCursorPasFg = pasFg;
  mCursorPasBg = pasBg;
}


void CHexBuffer::setSeparatorColor( const QColor &left, const QColor &right )
{
  mLeftSeparatorFg  = left;
  mRightSeparatorFg = right;
}

void CHexBuffer::setOffsetColor( const QColor &color )
{
  mTextOffsetFg = color;
}

void CHexBuffer::setNonPrintChar( char nonPrintChar )
{
  mNonPrintChar = nonPrintChar;
}

void CHexBuffer::setShowCursor( bool showCursor )
{
  mShowCursor = showCursor;
}


void CHexBuffer::setDisableCursor( bool disableCursor )
{
  mDisableCursor = disableCursor;
}


void CHexBuffer::setEditMode( EEditMode editMode )
{
  mEditMode = editMode;
  if( mEditMode == edit_insert )
  {
    if( mActiveEditArea == edit_primary )
    {
      mCursor.setShape( SCursorSpec::thin, SCursorSpec::frame, mUnitWidth,
			mNumCell );
    }
    else
    {
      mCursor.setShape( SCursorSpec::frame, SCursorSpec::thin, mUnitWidth,
			mNumCell );
    }

  }
  else
  {
    if( mActiveEditArea == edit_primary )
    {
      mCursor.setShape( SCursorSpec::solid, SCursorSpec::frame, mUnitWidth,
			mNumCell );
    }
    else
    {
      mCursor.setShape( SCursorSpec::frame, SCursorSpec::solid, mUnitWidth,
			mNumCell );
    }
  }
}


bool CHexBuffer::saveFile( QFile &file ) 
{
  int writeSize = file.writeBlock( data(), mDocumentSize ); 
  if( writeSize == -1 )
  {
    QString msg( i18n("Could not write file.") );
    QMessageBox::warning(
      0,
      i18n("Sorry"),
      msg.data(),
      i18n("OK"),
      "",
      "",
      0,0);
    return( false );
  }

  mDocumentModified = false;
  return( true );
}



bool CHexBuffer::loadFile( QFile *file, const QString &url ) 
{
  if( file == 0 )
  {
    if( resize( 100 ) == 0 )
    {
      QString msg;
      msg += i18n("Could not create new file." ); 
      msg += "\n";
      msg += i18n( "Insufficient memory." );
      QMessageBox::warning(
        0,
	i18n("Sorry"),
	msg.data(),
	i18n("OK"),
	"",
	"",
	0,0);
      return( false );
    }
    
    mDocumentSize = 0;
  }
  else
  {
    if( resize( file->size() + 100 ) == false )
    {
      QString msg( i18n("Could not read file.") ); 
      msg += "\n";
      msg += i18n( "Insufficient memory." );

      QMessageBox::warning(
        0,
	i18n("Sorry"),
	msg.data(),
	i18n("OK"),
	"",
	"",
	0,0);
      return( false );
    }

    int readSize = file->readBlock( data(), file->size() ); 
    if( readSize == -1 )
    {
      QString msg( i18n("Could not read file.") );
      QMessageBox::warning(
        0,
	i18n("Sorry"),
	msg.data(),
	i18n("OK"),
	"",
	"",
	0,0 );
      return( false );
    }

    mDocumentSize = file->size();
  }
  
  setUrl( url );
  mDocumentModified = false;

  computeNumLines();
  selectionSet( 0, true );

  return( true );
}




bool CHexBuffer::insertFile( QFile *file, const QString &url )
{
  if( file == 0 || file->size() == 0 )
  {
    return( false );
  }

  QByteArray array( file->size() );
  if( array.isNull() == true )
  {
    return( false );
  }

  int readSize = file->readBlock( array.data(), file->size() ); 
  if( readSize == -1 )
  {
    return( false );
  }

  bool success = inputAtCursor( array.data(), array.size(), 0 );
  return( success );
}





void CHexBuffer::closeFile( void ) 
{
  resize(0);
  mDocumentSize = 0;
  mDocumentModified = false;
  
  QString emptyUrl;
  setUrl( emptyUrl );

  computeNumLines();
  selectionSet( 0, true );
}


void CHexBuffer::setFont( const char *family, int pointSize, int weight, 
			  bool italic )
{
  mFont.setFamily( family );
  mFont.setPointSize( pointSize );
  mFont.setWeight( pointSize );
  mFont.setItalic( italic );

  QFontMetrics fm( mFont );
  mFontHeight = fm.height();
  mFontAscent = fm.ascent();

  computeLineWidth();
}


void CHexBuffer::setFont( QFont &font )
{
  mFont = font;
  QFontMetrics fm( mFont );
  mFontHeight = fm.height();
  mFontAscent = fm.ascent();

  computeLineWidth();
}




void CHexBuffer::computeLineWidth( void )
{
  QFontMetrics fm( mFont );
  mUnitWidth  = fm.width( "M" );
  mSplitWidth = mDpyState.splitWidth == -1 ? mUnitWidth : mDpyState.splitWidth;
 
  if( mDpyState.offsetVisible == true )
  {
    if( mDpyState.offsetMode == SDisplayState::offset_decimal )
    {
      mOffsetSize = 10;
      printOffset = &CHexBuffer::printDecimalOffset;
    }
    else if( mDpyState.offsetMode == SDisplayState::offset_hexadecimal )
    {
      mOffsetSize = 9;
      if( mDpyState.offsetCaps == true )
      {
	printOffset = &CHexBuffer::printHexadecimalBigOffset;
      }
      else
      {
	printOffset = &CHexBuffer::printHexadecimalSmallOffset;
      }
    }
  }
  else
  {
    mOffsetSize = 0;
    printOffset = &CHexBuffer::printDummyOffset;
  }    

  if( mDpyState.viewMode == SDisplayState::view_hexadecimal )
  {
    mNumCell = 2;
    mCursor.setCellWeight( 4 );
    if( mDpyState.viewCaps == true )
    {
      printCell = &CHexBuffer::printHexadecimalBigCell;
      inputCell = &CHexBuffer::inputHexadecimal;
    }
    else
    {
      printCell = &CHexBuffer::printHexadecimalSmallCell;
      inputCell = &CHexBuffer::inputHexadecimal;
    }
  }
  else if( mDpyState.viewMode == SDisplayState::view_octal )
  {
    mNumCell  = 3;
    printCell = &CHexBuffer::printOctalCell;
    inputCell = &CHexBuffer::inputOctal;
    mCursor.setCellWeight( 3 );
  }
  else if( mDpyState.viewMode == SDisplayState::view_binary )
  {
    mNumCell  = 8;
    printCell = &CHexBuffer::printBinaryCell;
    inputCell = &CHexBuffer::inputBinary;
    mCursor.setCellWeight( 1 );
  }
  else if( mDpyState.viewMode == SDisplayState::view_ascii )
  {
    mNumCell  = 1;
    printCell = &CHexBuffer::printAsciiCell;
    inputCell = &CHexBuffer::inputAscii;
    mCursor.setCellWeight( 8 );
  }
  else
  {
    mNumCell  = 2;
    mDpyState.viewMode = SDisplayState::view_hexadecimal;
    mDpyState.viewCaps = false;
    printCell = &CHexBuffer::printHexadecimalSmallCell;
    inputCell = &CHexBuffer::inputHexadecimal;
    mCursor.setCellWeight( 4 );
  }

  //
  // 'mPrimaryWidth' is the number of pixels that are needed to display a 
  // line in the primary field. 
  //
  mPrimaryWidth = mDpyState.lineSize * mNumCell * mUnitWidth;

  if( mDpyState.splitGroup == true )
  {
    int numSplit = mDpyState.lineSize / mDpyState.groupSize;
    numSplit -= mDpyState.lineSize % mDpyState.groupSize == 0 ? 1 : 0;
    mPrimaryWidth += numSplit * mSplitWidth;
  }
  
  //
  // 'mSecondaryWidth' is the number of pixels that are needed to display a 
  // line in the secondary field (there are no spaces).
  //
  if( mDpyState.textMode == SDisplayState::text_hide )
  {
    mSecondaryWidth = 0;
  }
  else
  {
    mSecondaryWidth = mDpyState.lineSize * mUnitWidth;
  }

  //
  // 'mLineWidth' is the total number of pixels required to display 
  // offset data, separators, primary and secondary data on a line.
  //
  mLineWidth = mPrimaryWidth + mSecondaryWidth + mOffsetSize * mUnitWidth;

  //
  // The 'mFixedWidth' is the number of pixels of the width that stays the
  // same regardless of how many characters that are displayed.
  // This entity consists of the edge margins, the inner margins and the
  // separators.
  //
  mFixedWidth = mOffsetSize * mUnitWidth;
  
  //
  // The edge margin is always present in both ends.
  //
  mLineWidth  += mDpyState.edgeMarginWidth * 2;
  mFixedWidth += mDpyState.edgeMarginWidth * 2;

  //
  // 'mTextStart1' is the number of pixels from the left edge where the 
  // primary field starts.
  //
  mTextStart1 = mDpyState.edgeMarginWidth;
  if( mDpyState.offsetVisible == true )
  {
    int width;
    if( mDpyState.leftSeparatorWidth > 0 )
    {
      width = mDpyState.innerMarginWidth * 2 + mDpyState.leftSeparatorWidth;
    }
    else
    {
      width = (mDpyState.innerMarginWidth * 3) / 2;
    }

    mLineWidth  += width;
    mFixedWidth += width;
    mTextStart1 += width + mOffsetSize * mUnitWidth;
  }

  //
  // 'mTextStart2' is the number of pixels from the left edge where the 
  // secondary fields start.
  //
  mTextStart2 = mTextStart1;
  if( mDpyState.textMode != SDisplayState::text_hide )
  {
    int width;
    if( mDpyState.rightSeparatorWidth > 0 )
    {
      width = mDpyState.innerMarginWidth * 2 + mDpyState.rightSeparatorWidth;
    }
    else
    {
      width = (mDpyState.innerMarginWidth * 3) / 2;
    }

    mLineWidth  += width;
    mFixedWidth += width;
    mTextStart2 += width + mPrimaryWidth;
  }

  setEditMode( mEditMode );
  computeNumLines();
}


void CHexBuffer::computeNumLines( void )
{
  if( mDpyState.lineSize == 0 )
  {
    mNumLines = 1;
  }
  else
  {
    uint s = textSize() + 1;
    mNumLines = s / mDpyState.lineSize + (s % mDpyState.lineSize ? 1 : 0);
  }
}





void CHexBuffer::drawText( QPainter &paint, uint line, int sx, int x1, int x2 )
{
  uint fileOffset = line * mDpyState.lineSize;

  paint.fillRect( x1, 0, x2-x1, mFontHeight, mTextBg );
  if( insideSelection( fileOffset, mDpyState.lineSize ) == true )
  {   
    uint start = startSelection( fileOffset );
    uint stop  = stopSelection( fileOffset );
    uint width = stop - start;

    
    uint addStart, addWidth;
    addStart = (start / mDpyState.groupSize) * mSplitWidth;
    if( width == 0 )
    {
      addWidth = 0;
    }
    else
    {
      uint g = mDpyState.groupSize;
      addWidth = (((start % g) + width - 1) / g) * mSplitWidth;
    }
  
    int offset = mTextStart1 - sx;
    paint.fillRect( offset + start * mNumCell * mUnitWidth + addStart, 
		    0, width * mNumCell * mUnitWidth + addWidth,
		    mFontHeight, mSelectBg );

    if( mDpyState.textMode != SDisplayState::text_hide )
    {
      offset = mTextStart2 - sx;
      paint.fillRect( offset + start * mUnitWidth, 
		      0, width * mUnitWidth, 
		      mFontHeight, mSelectBg );
    }
  }


  uint dataSize;
  unsigned char *fileData;
  if( size() == 0 || fileOffset > textSize() )
  {
    if( size() == 0 )
    {
      return;
    }
    dataSize = 0;
    fileData = 0;
  }
  else
  {
    dataSize = textSize() - fileOffset;
    if( dataSize > mDpyState.lineSize ) { dataSize = mDpyState.lineSize; }
    fileData = (unsigned char*)&(data()[ fileOffset ]);
  }


  int offset = mDpyState.edgeMarginWidth - sx;
 
  if( mDpyState.offsetVisible == true )
  {
    int s1 = mOffsetSize * mUnitWidth;
    if( x1 < offset + s1 && x2 >= offset && fileData != 0 )
    {
      paint.setPen( mTextOffsetFg );
      THIS_FPTR(printOffset)( mPrintBuf, fileOffset );
      paint.drawText( offset, mFontAscent, mPrintBuf, mOffsetSize );
    }
    offset += s1;

    if( mDpyState.leftSeparatorWidth > 0 )
    {
      offset += mDpyState.innerMarginWidth;

      int s2 = mDpyState.leftSeparatorWidth + mDpyState.innerMarginWidth;
      if( x1 < offset + s2 && x2 >= offset )
      {
	QPen pen( mLeftSeparatorFg, mDpyState.leftSeparatorWidth );
	paint.setPen( pen );
	int center = offset + mDpyState.leftSeparatorWidth/2;
	paint.drawLine( center, 0, center, mFontHeight );
      }
      offset += s2;
    }
    else
    {
      offset += (mDpyState.innerMarginWidth * 3) / 2; 
    }
  }

  int localOffset = offset;
  for( uint i = 0; i < dataSize; i++ )
  {
    int s = mNumCell * mUnitWidth + 
      ((i+1) % mDpyState.groupSize == 0) * mSplitWidth;
    if( x1 < localOffset + s && x2 >= localOffset )
    {
      QColor &color = insideSelection( fileOffset+i ) == true ? mSelectFg :
	foregroundColor( i );
      paint.setPen( color );

      THIS_FPTR(printCell)( mPrintBuf, fileData[i] );
      paint.drawText( localOffset, mFontAscent, mPrintBuf, mNumCell );
    }
    localOffset += s;
  }

  offset += mPrimaryWidth; 
  if( mDpyState.textMode != SDisplayState::text_hide )
  {
    if( mDpyState.rightSeparatorWidth > 0 )
    {
      offset += mDpyState.innerMarginWidth;
      int s = mDpyState.innerMarginWidth + mDpyState.rightSeparatorWidth;
      if( x1 < offset + s && x2 >= offset )
      {
	QPen pen( mRightSeparatorFg, mDpyState.rightSeparatorWidth );
	paint.setPen( pen );
	int center = offset + mDpyState.rightSeparatorWidth/2;
	paint.drawLine( center, 0, center, mFontHeight );
      }
      offset += s;
    }
    else
    {
      offset += (mDpyState.innerMarginWidth * 3) / 2;
    }


    int s = mUnitWidth;
    for( uint i = 0; i < dataSize; i++ )
    {
      if( x1 < offset + s && x2 >= offset )
      {
	QColor &color = insideSelection( fileOffset+i ) == true ? 
	  mSelectFg : mTextSecondaryFg;
	paint.setPen( color );

	printAsciiCell( mPrintBuf, fileData[i] );
	paint.drawText( offset, mFontAscent, mPrintBuf, 1 );
      }
      offset += s;
    }
  }

  //
  // If the cursors are located on the line we have drawn we redraw 
  // them unless they have been disabled.
  //
  if( mDisableCursor == true )
  {
    return;
  }

  if( mCursor.curr.inside( fileOffset, fileOffset + mDpyState.lineSize ) )
  {
    drawCursor( paint, sx );
  }  

}




void CHexBuffer::drawCursor( QPainter &paint, int startx )
{
  if( size() == 0 )
  {
    return;
  }

  SCursorSpec &c = mCursor.curr;

  //
  // Draw the cursor in primary edit area. 
  //

  QColor bg, fg;
  if( insideSelection( c.offset ) == true )
  {
    bg = c.mPrimaryShape == SCursorSpec::solid ? mCursorActBg : mSelectBg;
    fg = mSelectFg; 
  }
  else
  {
    bg = c.mPrimaryShape == SCursorSpec::solid ? mCursorActBg : mTextBg;
    fg = foregroundColor( c.offset % mDpyState.lineSize );
  }

  if( c.mPrimaryShape == SCursorSpec::thin )
  {
    paint.fillRect( c.x1 - startx, 0, mUnitWidth, mFontHeight, bg );
    
    /*
    bool showCursor = true;
    if( mActiveEditArea == edit_primary )
    {
      showCursor = mShowCursor;
      paint.setPen( red );
    }
    else
    {
      paint.setPen( black );
    }
    */
    
    if( mActiveEditArea == edit_primary )
    {
      if( mShowCursor == true )
      {
	paint.setPen( red );
	int center = c.x1 - startx - 1;

	paint.drawLine( center, 0, center, mFontHeight - 1 );
	paint.drawLine( center-2, 0, center+2, 0 );
	paint.drawLine( center-2, mFontHeight-1, center+2, mFontHeight-1 );
      }
    }
    else
    {
      paint.setPen( mCursorActBg );
      paint.drawRect( c.x1 - startx, 0, mUnitWidth * mNumCell, mFontHeight );
    }

    /*
    bool showCursor = mActiveEditArea == edit_primary ? mShowCursor : true;
    if( showCursor == true )
    {
      paint.setPen( red );
      int center = c.x1 - startx - 1;

      paint.drawLine( center, 0, center, mFontHeight - 1 );
      paint.drawLine( center-2, 0, center+2, 0 );
      paint.drawLine( center-2, mFontHeight-1, center+2, mFontHeight-1 );
    }
    */
  }
  else
  {
    if( mActiveEditArea == edit_primary )
    {
      if( mShowCursor == true )
      {
	paint.fillRect( c.x1 - startx, 0, mUnitWidth, mFontHeight, bg );
      }
    }
    else
    {
      paint.setPen( mCursorActBg );
      paint.drawRect( c.x1 - startx, 0, mUnitWidth*mNumCell, mFontHeight );
    }
  }

  if( c.offset < textSize() )
  {
    paint.setPen( fg );
    THIS_FPTR(printCell)( mPrintBuf, (unsigned char)c.data );
    paint.drawText( c.x1 - startx, mFontAscent, &mPrintBuf[c.cell], 1 );
  }
  
  //
  // Draw the cursor in secodary edit area. 
  //

  if( mDpyState.textMode != SDisplayState::text_hide )
  {
    QColor bg, fg;
    if( insideSelection( c.offset ) == true )
    {
      bg = c.mSecondaryShape == SCursorSpec::solid ? mCursorActBg : mSelectBg;
      fg = mSelectFg;
    }
    else
    {
      bg = c.mSecondaryShape == SCursorSpec::solid ? mCursorActBg : mTextBg;
      fg = mTextSecondaryFg;
    }

    if( c.mSecondaryShape == SCursorSpec::thin )
    {
      paint.fillRect( c.x2 - startx, 0, mUnitWidth, mFontHeight, bg );

      /*
      if( mActiveEditArea == edit_secondary )
      {
	if( mShowCursor == false )
	{
	  return;
	}
	paint.setPen( red );
      }
      else
      {
	paint.setPen( black );
      }
      */

      if( mActiveEditArea == edit_secondary )
      {
	if( mShowCursor == true )
	{
	  paint.setPen( red );
	  int center = c.x2 - startx - 1;

	  paint.drawLine( center, 0, center, mFontHeight - 1 );
	  paint.drawLine( center-2, 0, center+2, 0 );
	  paint.drawLine( center-2, mFontHeight-1, center+2, mFontHeight-1 );
	}
      }
      else
      {
	paint.setPen( mCursorActBg );
	paint.drawRect( c.x2 - startx, 0, mUnitWidth, mFontHeight );
      }

      /*
      bool showCursor = mActiveEditArea == edit_secondary ? mShowCursor : true;
      if( showCursor == true )
      {
	paint.setPen( red );
	int center = c.x2 - startx - 1;
	paint.drawLine( center, 0, center, mFontHeight - 1 );
	paint.drawLine( center-2, 0, center+2, 0 );
	paint.drawLine( center-2, mFontHeight-1, center+2, mFontHeight-1 );
      }
      */
    }
    else
    {
      if( mActiveEditArea == edit_secondary )
      {
	if( mShowCursor == true )
	{
	  paint.fillRect( c.x2 - startx, 0, mUnitWidth, mFontHeight, bg );
	}
      }
      else
      {
	paint.setPen( mCursorActBg );
	paint.drawRect( c.x2 - startx, 0, mUnitWidth, mFontHeight );
      }
    }
    
    if( c.offset < textSize() )
    {
      paint.setPen( fg );
      printAsciiCell( mPrintBuf, c.data );
      paint.drawText( c.x2 - startx, mFontAscent, mPrintBuf, 1 );
    }
  }
}



void CHexBuffer::cursorReset( void )
{
  mCursor.reset();
  cursorCompute();
}

void CHexBuffer::cursorCompute( void )
{
  mCursor.prev = mCursor.curr;

  if( mCursor.next.offset >= textSize() ) 
  {
    if( textSize() == 0 )
    {
      mCursor.curr.offset  = 0;
      mCursor.curr.data    = 0;
      mCursor.curr.cell    = 0;
      mCursor.curr.maxCell = mNumCell;

      int x = mCursor.curr.offset % mDpyState.lineSize;
      mCursor.curr.x1 = mTextStart1;
      mCursor.curr.x1 += (x * mNumCell + mCursor.curr.cell) * mUnitWidth;
      mCursor.curr.x1 += (x / mDpyState.groupSize) * mSplitWidth;
      mCursor.curr.x2 = mTextStart2 + x * mUnitWidth;

      /*
      mCursor.curr.x1 = mTextStart1 + 
	(x*mNumCell + x/mDpyState.groupSize + mCursor.curr.cell) * mUnitWidth;
      mCursor.curr.x2 = mTextStart2 + mUnitWidth * x;
      */
      mCursor.curr.y  = (mCursor.curr.offset/mDpyState.lineSize) * mFontHeight;
      return;

    }
    mCursor.next.offset = textSize();
  }

  mCursor.curr.offset  = mCursor.next.offset;
  mCursor.curr.data    = data()[ mCursor.curr.offset ];
  mCursor.curr.cell    = mCursor.next.cell;
  mCursor.curr.maxCell = mNumCell;

  int x = mCursor.curr.offset % mDpyState.lineSize;

  mCursor.curr.x1 = mTextStart1;
  mCursor.curr.x1 += (x * mNumCell + mCursor.curr.cell) * mUnitWidth;
  mCursor.curr.x1 += (x / mDpyState.groupSize) * mSplitWidth;
  mCursor.curr.x2 = mTextStart2 + x * mUnitWidth;

  //mCursor.curr.x1 = mTextStart1 + 
  //(x*mNumCell + x/mDpyState.groupSize + mCursor.curr.cell) * mUnitWidth;
  //mCursor.curr.x2 = mTextStart2 + mUnitWidth * x;
  mCursor.curr.y  = (mCursor.curr.offset/mDpyState.lineSize) * mFontHeight; 
} 


int CHexBuffer::setCursor( int x, int y, bool init )
{
  if( size() == 0 )
  {
    return( edit_none );
  }


  uint line  = y < 0 ? 0 : y / mFontHeight;
  uint entry = 0;

  if( init == false )
  {
    if( mCursor.area() == edit_primary )
    {
      int start = mTextStart1;
      if( x < start - mDpyState.innerMarginWidth )
      {
	return( edit_none );
      }
      else
      {
	int stop  = mTextStart1 + mPrimaryWidth + mDpyState.innerMarginWidth;
	int width = mNumCell * mUnitWidth;
	int space = mSplitWidth;//mUnitWidth;
      
	for( int position = start, i=0; position < stop; i++ )
	{
	  if( x <= position + width )
	  {
	    break;
	  }
	  position += width + (((i+1) % mDpyState.groupSize) ? 0 : space);
	  entry    += 1;
	}
      }
    }
    else 
    {
      int start = mTextStart2;
      if( x < start - mDpyState.innerMarginWidth || 
	  mDpyState.textMode == SDisplayState::text_hide )
      {
	return( 0 );
      }
      int stop  = mTextStart2 + mDpyState.lineSize * mUnitWidth;
      int width = mUnitWidth * 1;
      int space = 0;

      for( int position = start; position < stop; )
      {
	if( x <= position + width )
	{
	  break;
	}
	position += width + space;
	entry    += 1;
      }
    }
  }
  else
  {
    int start = mTextStart1;
    int stop  = start + mPrimaryWidth + mDpyState.innerMarginWidth;
    if( x >= start - mDpyState.innerMarginWidth && x <= stop )
    {
      int width = mUnitWidth * mNumCell;
      int space = mSplitWidth;//mUnitWidth;
   
      for( int position = start, i=0; position < stop; i++ )
      {
	if( x <= position + width )
	{
	  break;
	}
	position += width + (((i+1) % mDpyState.groupSize) ? 0 : space);
	entry    += 1;
      }

      mActiveEditArea = edit_primary;
    }
    else if( mDpyState.textMode != SDisplayState::text_hide )
    {
      start = mTextStart2;
      stop  = mTextStart2 + mDpyState.lineSize * mUnitWidth + 
	mDpyState.edgeMarginWidth;
      if( x >= start - mDpyState.innerMarginWidth && x <= stop )
      {
	int width = mUnitWidth * 1;
	int space = 0;

	for( int position = start; position < stop; )
	{
	  if( x <= position + width )
	  {
	    break;
	  }
	  position += width + space;
	  entry    += 1;
	}

	mActiveEditArea = edit_secondary;
      }
    }
    else
    {
      return( edit_none );
    }
  }

  uint offset = line * mDpyState.lineSize + entry;
  if( offset > textSize() ) 
  {  
    offset = textSize(); 
  }

  mCursor.setOffset( offset );
  cursorCompute();
  if( mActiveEditArea != mCursor.area() )
  {
    mCursor.setArea( mActiveEditArea );
    setEditMode( mEditMode );
  }

  return( mCursor.area() );
}




bool CHexBuffer::inputAtCursor( int ascii )
{
  if( size() == 0 || mActiveEditArea == edit_none )
  {
    return( false );
  }

  unsigned char dest;
  bool insert;
  if( mEditMode == edit_overwrite || mCursor.curr.cell > 0 )
  {
    if( mCursor.curr.offset >= mDocumentSize )
    {
      dest   = 0;
      insert = true;
    }
    else
    {
      dest   = (unsigned char)data()[ mCursor.curr.offset ];
      insert = false;
    }
  }
  else
  {
    dest   = 0;
    insert = true;
  }


  if( mActiveEditArea == edit_primary )
  {
    if( THIS_FPTR(inputCell)( &dest, ascii, mCursor.curr.cell ) == false )
    {
      return( false );
    }
  }
  else if( mActiveEditArea == edit_secondary )
  {
    if( inputAscii( &dest, ascii, mCursor.curr.cell ) == false )
    {
      return( false );
    }
  }
  else 
  {
    return( false );
  }

  recordStart( mCursor );
  recordReplace( mCursor, insert == true ? 0 : 1, (char*)&dest, 1 );
  cursorRight( cursorPrimaryEdit() );
  recordEnd( mCursor );

  computeNumLines();
  return( true );
}


bool CHexBuffer::inputAtCursor( const char *buf, uint bufSize, uint oldSize )
{
  if( size() == 0 || buf == 0 || bufSize == 0 )
  {
    return( false );
  }

  recordStart( mCursor );
  recordReplace( mCursor, oldSize, (char*)buf, bufSize );
  cursorStep( bufSize, true, false );
  recordEnd( mCursor );

  computeNumLines();
  return( true );
}





bool CHexBuffer::removeAtCursor( bool beforeCursor )
{
  if( size() == 0 )
  {
    return( false );
  }

  if( selectionValid() == true )
  {
    cutSelection();
    return( true );
  }


  if( beforeCursor == true )
  {
    if( mCursor.curr.offset == 0 )
    {
      return( false );
    }

    recordStart( mCursor );
    cursorLeft( false );
    recordReplace( mCursor, 1, 0, 0 );
    recordEnd( mCursor );

    computeNumLines();
    return( true );
  }
  else
  {
    if( mCursor.curr.offset + 1 > mDocumentSize )
    {
      return( false );
    }
    
    recordStart( mCursor );
    recordReplace( mCursor, 1, 0, 0 );
    recordEnd( mCursor );

    computeNumLines();
    return( true );
  }
}



bool CHexBuffer::selectedData( QByteArray &array, bool formatted )
{
  puts("CHexBuffer::selectedData: 1");

  uint start = mSelect.curr.start;
  uint stop  = mSelect.curr.stop;

  if( selectionValid() == false || start >= stop )
  {
    return( false );
  }
  
  uint size;
  if( formatted == true )
  {
    size = 1 + (stop - start + 1) * mNumCell;
  }
  else
  {
    printf("CHexBuffer::selectedData: start=%d, stop=%d\n", stop, start );
    size = stop - start;
  }

  if( array.resize( size ) == false )
  {
    return( false );
  }

  puts("CHexBuffer::selectedData: A");

  unsigned char *src = (unsigned char*)data();
  char *dst = (char*)array.data();

  if( formatted == true )
  {
    for( uint i=start; i<=stop; i++ )
    {
      THIS_FPTR(printCell)( &dst[(i-start)*mNumCell], src[i] );
    }
    dst[size] = 0;
  }
  else
  {
    memcpy( dst, &src[start], size );
  }

  puts("CHexBuffer::selectedData: B");

  return( true );
}


bool CHexBuffer::export( QByteArray &array, SExportData &ed )
{
  uint start, stop;
  if( ed.mode == SExportData::Everything )
  {
    start = 0;
    stop  = textSize();
  }
  else if( ed.mode == SExportData::SelectedArea )
  {
    if( selectionValid() == false )
    {
      return( false );
    }
    start = mSelect.curr.start;
    stop  = mSelect.curr.stop;
  }
  else if( ed.mode == SExportData::Specified )
  {
    start = ed.start;
    stop  = ed.stop;
  }
  else
  {
    return( false );
  }
  
  if( start >= stop )
  {
    return( false );
  }

  uint startLine = calculateLine( start );
  uint stopLine  = calculateLine( stop );

  if( startLine >= (uint)numLines() )
  {
    startLine = numLines() == 0 ? 0 : numLines() - 1;
  }
  if( stopLine >= (uint)numLines() )
  {
    stopLine = numLines() == 0 ? 0 : numLines() - 1;
  }

  uint bytePerLine = mOffsetSize + 1 + (mNumCell + 2)*mDpyState.lineSize + 1;
  uint size = (stopLine - startLine + 1)*bytePerLine;

  if( array.resize( size+1 ) == false )
  {
    return( false );
  }

  char *dst = (char*)array.data();
  for( uint i = startLine; i <= stopLine; i++ )
  {
    uint printSize = printLine( dst, i );
    dst += printSize;
  }
  *dst = 0;

  return( true );
}






bool CHexBuffer::exportData( QByteArray &array, uint startLine, uint stopLine )
{
  if( stopLine >= (uint)numLines() )
  {
    stopLine = numLines() == 0 ? 0 : numLines() - 1;
  }

  if( startLine > stopLine )
  {
    startLine = stopLine;
  }
  
  uint bytePerLine = mOffsetSize + 1 + (mNumCell + 2)*mDpyState.lineSize + 1;
  uint size = (stopLine - startLine + 1)*bytePerLine;

  if( array.resize( size+1 ) == false )
  {
    return( false );
  }

  char *dst = (char*)array.data();
  for( uint i = startLine; i <= stopLine; i++ )
  {
    uint printSize = printLine( dst, i );
    dst += printSize;
  }
  *dst = 0;

  return( true );
}



uint CHexBuffer::printLine( char *dst, uint line )
{
  uint offset = line * mDpyState.lineSize;
  unsigned char *src;
  char *start = dst;

  uint dataSize;
  if( offset >= textSize() )
  {
    src = 0;
    dataSize = 0;
  }
  else
  {
    src = (unsigned char*)&data()[ offset ];
    dataSize = textSize() - offset;
  }

  if( mDpyState.offsetVisible == true )
  {
    THIS_FPTR(printOffset)( dst, offset ); dst += mOffsetSize;
    sprintf( dst, " " ); dst += 1;
  }
  for( uint i=0; i<mDpyState.lineSize; i++ )
  {
    if( i<dataSize )
    { 
      THIS_FPTR(printCell)( dst, src[i] ); dst += mNumCell;
    }
    else
    {
      memset( dst, ' ', mNumCell ); dst += mNumCell;
    }
    if( mDpyState.splitGroup == true )
    {
      sprintf( dst, " " ); dst += 1;
    }
  }
  if( mDpyState.textMode != SDisplayState::text_hide )
  {
    for( uint i=0; i<mDpyState.lineSize; i++ )
    {
      if( i<dataSize )
      {
	printAsciiCell( dst, src[i] ); dst += 1;
      }
      else
      {
	memset( dst, ' ', 1 ); dst += 1;
      }
    }
  }
  sprintf( dst, "\n" ); dst += 1;
  return( (uint)(dst-start) );
}














bool CHexBuffer::cutSelection( void )
{
  if( size() == 0 || selectionValid() == false )
  {
    return( false );
  }

  uint start = startSelection();
  uint stop  = stopSelection();
  selectionSet( 0, true );
    
  recordStart( mCursor );
  cursorGoto( start, 0 );
  recordReplace( mCursor, stop-start, 0, 0 );
  recordEnd( mCursor );

  computeNumLines();
  return( true );
}



bool CHexBuffer::undo( void )
{
  if( size() == 0 || mUndoIndex == 0 )
  {
    return( false );
  }
  
  CHexActionGroup *group = mUndoList.at( mUndoIndex -1 );
  if( group == 0 )
  {
    return( false );
  }

  mUndoIndex -= 1;
  doActionGroup( group );

  cursorGoto( group->mStartOffset, group->mStartBit );

  return( true );
}


bool CHexBuffer::redo( void )
{
  if( size() == 0 || mUndoIndex >= mUndoList.count() )
  {
    return( false );
  }

  CHexActionGroup *group = mUndoList.at( mUndoIndex );
  if( group == 0 )
  {
    return( false );
  }

  mUndoIndex += 1;
  doActionGroup( group );

  cursorGoto( group->mStopOffset, group->mStopBit );

  return( true );
}



bool CHexBuffer::search( USearchData &sd, bool replaceData )
{
  if( sd.mAny.operation == USearchData::opGoto )
  {
    SSearchGoto &g = sd.mGoto;
    cursorGoto( g.offset, g.bit, g.backward, g.fromCursor );
    return( true );
  }
  else if( sd.mAny.operation == USearchData::opFind )
  {
    SSearchFind &f = sd.mFind;
    if( f.pattern == 0 || f.size == 0 || f.size > mDocumentSize )
    {
      return( false );
    }

    uint start;
    if( f.fromCursor == true )
    {
      if( f.backward == true )
      {
	start = cursorOffset() - (cursorOffset() > 0 && f.step==true ? 1 : 0);
      }
      else
      {
	start = cursorOffset() + (f.step == true ? 1 : 0);
      }
    }
    else
    {
      if( f.backward == true )
      {
	start = mDocumentSize - (mDocumentSize > 0 && f.step == true ? 1 : 0);
      }
      else
      {
	start = f.step == true ? 1 : 0;
      }
    }

    if( f.backward == true )
    {
      uint stop = 0;
      if( start + f.size > mDocumentSize )
      {
	start = mDocumentSize < f.size ? 0 : mDocumentSize - f.size;
      }

      if( f.inSelection == true )
      {
	if( selectionValid() == false )
	{
	  return( false );
	}
	else
	{
	  if( stopSelection() < start ) { start = stopSelection(); }
	  if( startSelection() > stop ) { stop = startSelection(); }
	  if( stop + f.size > start ) { return( false ); }
	}
      }

      for( uint i = start; i > stop; i-- )
      {
	if( memcmp( &data()[i], f.pattern, f.size ) == 0 )
	{
	  cursorGoto( i, 7 );
	  return( true );
	}
      }
    }
    else
    {
      uint stop = mDocumentSize;
      if( f.inSelection == true )
      {
	if( selectionValid() == false )
	{
	  return( false );
	}
	else
	{
	  if( startSelection() > start ) { start = startSelection(); }
	  if( stopSelection() < stop ) { stop = stopSelection(); }
	  if( stop < f.size ) { return( false ); }
	}
      }

      for( uint i = start; i < stop - f.size; i++ )
      {
	if( memcmp( &data()[i], f.pattern, f.size ) == 0 )
	{
	  cursorGoto( i, 7 );
	  return( true );
	}
      }
    }
    return( false );

  }
  else if( sd.mAny.operation == USearchData::opReplace )
  {
    SSearchReplace &r = sd.mReplace;
    if( r.pattern == 0 || r.pSize == 0 || r.pSize > mDocumentSize )
    {
      return( false );
    }

    uint start;
    if( r.fromCursor == true )
    {
      if( r.backward == true )
      {
	start = cursorOffset() - (cursorOffset() > 0 && r.step==true ? 1 : 0);
      }
      else
      {
	start = cursorOffset() + (r.step == true ? 1 : 0);
      }
    }
    else
    {
      if( r.backward == true )
      {
	start = mDocumentSize - (mDocumentSize > 0 && r.step == true ? 1 : 0);
      }
      else
      {
	start = r.step == true ? 1 : 0;
      }
    }

    if( r.backward == true )
    {
      uint stop = 0;
      if( start + r.pSize > mDocumentSize )
      {
	start = mDocumentSize < r.pSize ? 0 : mDocumentSize - r.pSize;
      }

      if( r.inSelection == true )
      {
	if( selectionValid() == false )
	{
	  return( false );
	}
	else
	{
	  if( stopSelection() < start ) { start = stopSelection(); }
	  if( startSelection() > stop ) { stop = startSelection(); }
	  if( stop > start ) { return( false ); }
	}
      }

      for( uint i = start; i >= stop; i-- )
      {
	if( memcmp( &data()[i], r.pattern, r.pSize ) == 0 )
	{
	  cursorGoto( i, 7 );
	  if( replaceData == true )
	  {
	    inputAtCursor( r.replacement, r.rSize, r.pSize );
	    if( r.inSelection == true )
	    {
	      if( r.pSize > r.rSize )
	      {
		mSelect.shrink( r.pSize - r.rSize );
	      }
	      else
	      {
		mSelect.expand( r.rSize - r.pSize );
	      }
	    }
	  }
	  return( true );
	}
      }
    }
    else
    {
      uint stop = mDocumentSize;
      if( r.inSelection == true )
      {
	if( selectionValid() == false )
	{
	  return( false );
	}
	else
	{
	  if( startSelection() > start ) { start = startSelection(); }
	  if( stopSelection() < stop ) { stop = stopSelection(); }
	  if( stop < r.pSize ) { return( false ); }
	}
      }

      for( uint i = start; i <= stop - r.pSize; i++ )
      {
	if( memcmp( &data()[i], r.pattern, r.pSize ) == 0 )
	{
	  cursorGoto( i, 7 );
	  if( replaceData == true )
	  {
	    inputAtCursor( r.replacement, r.rSize, r.pSize );
	    if( r.inSelection == true )
	    {
	      if( r.pSize > r.rSize )
	      {
		mSelect.shrink( r.pSize - r.rSize );
	      }
	      else
	      {
		mSelect.expand( r.rSize - r.pSize );
	      }
	    }
	  }
	  return( true );
	}	
      }
    }
    return( false );

  }
  return( false );
}


uint CHexBuffer::replaceAll( USearchData &sd )
{
  if( sd.mAny.operation != USearchData::opReplace )
  {
    return( 0 );
  }

  SSearchReplace &r = sd.mReplace;
  if( r.pattern == 0 || r.pSize == 0 || r.pSize > mDocumentSize )
  {
    return( 0 );
  }

  uint numReplaced = 0;
  uint start;

  if( r.fromCursor == true )
  {
    start = cursorOffset();
  }
  else
  {
    start = r.backward == true ? mDocumentSize : 0;
  }

  recordStart( mCursor );

  if( r.backward == true )
  {
    uint stop = 0;
    if( start + r.pSize > mDocumentSize )
    {
      start = mDocumentSize < r.pSize ? 0 : mDocumentSize - r.pSize;
    }

    if( r.inSelection == true )
    {
      if( selectionValid() == false )
      {
	return( false );
      }
      else
      {
	if( stopSelection() < start ) { start = stopSelection(); }
	if( startSelection() > stop ) { stop = startSelection(); }
	if( stop > start ) { return( false ); }
      }
    }

    for( uint i = start; i >= stop; i-- )
    {
      if( memcmp( &data()[i], r.pattern, r.pSize ) == 0 )
      {
	cursorGoto( i, 7 );
	recordReplace( mCursor, r.pSize, (char*)r.replacement, r.rSize );
	if( r.inSelection == true )
	{
	  if( r.pSize > r.rSize )
	  {
	    mSelect.shrink( r.pSize - r.rSize );
	  }
	  else
	  {
	    mSelect.expand( r.rSize - r.pSize );
	  }
	}
	cursorStep( r.rSize, true, false );
	numReplaced += 1;
      }
    }
  }
  else
  {
    uint stop = mDocumentSize;
    if( r.inSelection == true )
    {
      if( selectionValid() == false )
      {
	return( false );
      }
      else
      {
	if( startSelection() > start ) { start = startSelection(); }
	if( stopSelection() < stop ) { stop = stopSelection(); }
	if( stop < r.pSize ) { return( false ); }
      }
    }

    for( uint i = start; i <= stop - r.pSize; i++ )
    {
      if( memcmp( &data()[i], r.pattern, r.pSize ) == 0 )
      {
	cursorGoto( i, 7 );
	recordReplace( mCursor, r.pSize, (char*)r.replacement, r.rSize );
	if( r.inSelection == true )
	{
	  if( r.pSize > r.rSize )
	  {
	    mSelect.shrink( r.pSize - r.rSize );
	    uint diff = r.pSize - r.rSize;
	    if( stop - r.pSize < diff ) { break; }
	    stop -= diff;
	  }
	  else
	  {
	    mSelect.expand( r.rSize - r.pSize );
	    stop += r.rSize - r.pSize;
	  }
	}

	cursorStep( r.rSize, true, false );
	numReplaced += 1;
      }
    }
  }

  recordEnd( mCursor );
  computeNumLines();

  return( numReplaced );
}




void CHexBuffer::doActionGroup( CHexActionGroup *group )
{
  if( group == 0 )
  {
    return;
  }
  
  CHexAction *action = group->mHexAction;
  group->mHexAction = 0;

  while( action != 0 )
  {
    doAction( action );
    CHexAction *next = action->mNext;
    group->insertAction( action );
    action = next;
  }

  computeNumLines();
}


void CHexBuffer::doAction( CHexAction *action ) 
{
  if( action->mAction == CHexAction::replace )
  {
    doReplace( action, true );
  }
}





void CHexBuffer::recordStart( SCursor &cursor )
{
  //
  // Step 1: Remove any undo element that is more recent than the 
  // current undo index
  //
  while( mUndoList.count() > mUndoIndex ) 
  {
    mUndoList.removeLast();
  }

  //
  // Step 2: Make sure the undo list is no larger than the undo limit.
  // We remove the oldest elements in the list.
  //
  while( mUndoList.count() >= mUndoLimit ) 
  {
    mUndoList.removeFirst();
    mUndoIndex -= 1;
  }

  CHexActionGroup *group = new CHexActionGroup( cursor.curr.offset,
						cursor.bit() );
  if( group == 0 )
  {
    return;
  }

  mUndoList.append( group );
  mUndoIndex += 1;
}


void CHexBuffer::recordReplace( SCursor &cursor, uint size, char *data1, 
				uint data1Size )
{
  CHexAction *hexAction = new CHexAction( CHexAction::replace, 
					  cursor.curr.offset );
  if( hexAction == 0 )
  {
    return;
  }

  hexAction->mSize = size;
  hexAction->mData = data1;
  hexAction->mDataSize = data1Size;

  doReplace( hexAction, false );
  mUndoList.getLast()->insertAction( hexAction );

  if( mCursor.curr.offset < mDocumentSize )
  {
    mCursor.curr.data = data()[ mCursor.curr.offset ];
  }

}

void CHexBuffer::recordEnd( SCursor &cursor )
{
  mUndoList.getLast()->mStopOffset = cursor.curr.offset;
  mUndoList.getLast()->mStopBit = cursor.bit();
}


void CHexBuffer::doReplace( CHexAction *hexAction, bool removeData )
{
  //
  // Mark data as changed. Every data altering operation results in 
  // this function being executed. So this ia a convenient place
  //
  mDocumentModified = true;

  uint offset   = hexAction->mOffset;
  uint oldSize  = hexAction->mSize;
  char *newData = hexAction->mData;
  uint newSize  = hexAction->mDataSize;

  hexAction->setData( newSize, &data()[offset], oldSize );

  //
  // Input new data. Resize buffer first.
  //
  bool success;
  if( newSize > oldSize )
  {
    success = moveBuffer( offset + newSize - oldSize, offset );
  }
  else
  {
    success = moveBuffer( offset, offset + oldSize - newSize );
  }

  if( success == true )
  {
    memcpy( &data()[offset], newData, newSize );
  }

  if( removeData == true )
  {
    delete [] newData;
  }

}











bool CHexBuffer::inputDummy( unsigned char *dest, int value, uint cell )
{
  return( false );
}


bool CHexBuffer::inputHexadecimal( unsigned char *dest, int value, uint cell )
{ 
  if( value >= '0' && value <= '9' )
  {
    value = value - '0';
  }
  else if( value >= 'A' && value <= 'F' )
  {
    value = value - 'A' + 10;
  }
  else if( value >= 'a' && value <= 'f' )
  {
    value = value - 'a' + 10;
  }
  else
  {
    return( false );
  }

  if( cell > 1 )
  {
    return( false );
  }

  uint shift = 1 - cell;
  *dest = (*dest & ~(0xF<<(shift*4)) ) | (value<<(shift*4));
  return( true );
}




bool CHexBuffer::inputOctal( unsigned char *dest, int value, uint cell )
{ 
  if( value >= '0' && value <= '7' )
  {
    value = value - '0';
    if( cell == 0 && value > 3 )
    {
      return( false );
    }
  }
  else
  {
    return( false );
  }

  if( cell > 3 )
  {
    return( false );
  }

  uint shift = 2 - cell;
  *dest = (*dest & ~(0x7<<(shift*3)) ) | (value<<(shift*3));
  return( true );
}


bool CHexBuffer::inputBinary( unsigned char *dest, int value, uint cell )
{ 
  if( value >= '0' && value <= '1' )
  {
    value = value - '0';
  }
  else
  {
    return( false );
  }

  if( cell > 7 )
  {
    return( false );
  }

  uint shift = 7 - cell;
  *dest = (*dest & ~(1<<shift)) | (value<<shift);
  return( true );
}



bool CHexBuffer::inputAscii( unsigned char *dest, int value, uint cell )
{ 
  *dest = value;
  return( true );
}







bool CHexBuffer::incTextSize( uint offset, uint incSize, bool insert )
{
  if( insert == false )
  {
    if( resizeBuffer( offset + incSize ) == false )
    {
      return( false );
    }
  }
  else
  {
    if( moveBuffer( offset + incSize, offset ) == false )
    {
      return( false );
    }
  }

  return( true );
}



bool CHexBuffer::moveBuffer( uint destOffset, uint srcOffset )
{
  if( srcOffset > mDocumentSize || destOffset == srcOffset )
  {
    return( true );
  }

  if( destOffset < srcOffset )
  {
    char *dest = &data()[ destOffset ];
    char *src  = &data()[ srcOffset ];
    
    memmove( dest, src, mDocumentSize - srcOffset );
    mDocumentSize -= (srcOffset - destOffset);
    return( true );
  }
  else
  {
    uint s = mDocumentSize - srcOffset;
    if( destOffset + s >= size() )
    {
      if( resizeBuffer( destOffset + s ) == false )
      {
	return( false );
      }
    }
    else
    {
      mDocumentSize += (destOffset - srcOffset);
    }

    char *dest = &data()[ destOffset ];
    char *src  = &data()[ srcOffset ];

    memmove( dest, src, s );
    memset( src, 0, destOffset - srcOffset ); 
    return( true );
  }
}




bool CHexBuffer::resizeBuffer( uint newSize )
{
  if( newSize < textSize() )
  {
    return( true );
  }

  if( newSize >= size() )
  {
    QByteArray tmp;
    tmp.duplicate( data(), size() );
    if( tmp.isNull() == true )
    {
      QString msg( i18n("Could not enlarge file.") ); 
      msg += "\n";
      msg += i18n( "Insufficient memory." );

      QMessageBox::warning(
        0,
	i18n("Sorry"),
	msg.data(),
	i18n("OK"),
	"",
	"",
	0,0);
      return( false );
    }

    if( fill( '\0', newSize + 100 ) == false )
    {
      QString msg( i18n("Could not enlarge file.") ); 
      msg += "\n";
      msg += i18n( "Insufficient memory." );

      QMessageBox::warning(
        0,
	i18n("Sorry"),
	msg.data(),
	i18n("OK"),
	"",
	"",
	0,0);
      return( false );
    }
    memcpy( data(), tmp.data(), tmp.size() );
  }

  mDocumentSize = newSize;

  return( true );
}





